/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.config;

/**
 *
 * @author Philipp C. Heckel <philipp.heckel@gmail.com>
 */
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.NoSuchAlgorithmException;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.crypto.NoSuchPaddingException;
import org.syncany.Constants;
import org.syncany.exceptions.ConfigException;
import java.security.InvalidKeyException;
import java.security.MessageDigest;
import java.util.Arrays;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.CipherOutputStream;
import javax.crypto.IllegalBlockSizeException;
import javax.crypto.spec.SecretKeySpec;
import org.syncany.exceptions.EncryptionException;

public class Encryption implements Configurable {
    private String password;
    private String cipherStr;
    private Integer keylength;

    private Cipher encCipher;
    private Cipher decCipher;

    public Encryption() {
        // All set by init()
        password = null;
        cipherStr = null;
        keylength = null;

        encCipher = null;
        decCipher = null;
    }

    /**
     * bit must be dividable by 8.
     *
     * @param password
     * @param bit
     */
    public void init() throws ConfigException {
        try {
            // Create the cipher object with the generated key
            this.encCipher = createEncCipher((byte[]) null); // unsalted.
            this.decCipher = createDecCipher((byte[]) null); // unsalted.
            
            // Check
            byte[] testBytes = new byte[] { 1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3, 4, 5, 6, 7, 8, 9 };

            if (!Arrays.equals(decrypt(encrypt(testBytes)), testBytes)) {
                throw new ConfigException("Test encrypt/decrypt cycle failed.");
            }
        }
        catch (Exception e) {
            throw new ConfigException("Invalid encryption parameters.", e);
        }
    }
    
    public Cipher createEncCipher(byte[] salt) throws EncryptionException {
        try {        
            SecretKeySpec keySpec = createKeySpec(salt);

            Cipher cipher = Cipher.getInstance(cipherStr);
            cipher.init(Cipher.ENCRYPT_MODE, keySpec);        

            return cipher;
        }
        catch (Exception e) {
            throw new EncryptionException(e);
        }
    }    
    
    public Cipher createEncCipher(String salt) throws EncryptionException {
        try {
            return createEncCipher(salt.getBytes("UTF-8"));
        } 
        catch (UnsupportedEncodingException ex) {
            throw new EncryptionException(ex);
        }
    }    
    
    public Cipher createDecCipher(byte[] salt) throws EncryptionException {
        try {
            SecretKeySpec keySpec = createKeySpec(salt);

            Cipher cipher = Cipher.getInstance(cipherStr);
            cipher.init(Cipher.DECRYPT_MODE, keySpec);        

            return cipher;
        }
        catch (Exception e) {
            throw new EncryptionException(e);
        }
    }    
    
    public Cipher createDecCipher(String salt) throws EncryptionException {
        try {
            return createDecCipher(salt.getBytes("UTF-8"));
        } 
        catch (UnsupportedEncodingException ex) {
            throw new EncryptionException(ex);
        }
    }    

    /**
     * unsalted.
     * @return 
     */
    public Cipher getEncCipher() {
        return encCipher;
    }
    
    /**
     * unsalted.
     * @return 
     */
    public Cipher getDecCipher() {
        return decCipher;
    }                

    public String getCipherStr() {
        return cipherStr;
    }

    public void setCipherStr(String cipherStr) {
        this.cipherStr = cipherStr;
    }

    public Integer getKeylength() {
        return keylength;
    }

    public void setKeylength(Integer keylength) {
        this.keylength = keylength;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    /**
     * unsalted.
     */
    public synchronized byte[] encrypt(byte[] data)
        throws InvalidKeyException, IllegalBlockSizeException, BadPaddingException {

        return encCipher.doFinal(data);
    }

    /**
     * unsalted.
     */
    public synchronized byte[] decrypt(byte[] data)
        throws InvalidKeyException, IllegalBlockSizeException, BadPaddingException {

        return decCipher.doFinal(data);
    }
    
    private SecretKeySpec createKeySpec(byte[] salt) 
            throws NoSuchAlgorithmException, UnsupportedEncodingException, EncryptionException {
        
        if (keylength % 8 != 0) {
            throw new EncryptionException("Invalid keylength. Must be divisible by 8.");
        }
        
        // Created salted password        
        byte[] saltedpass;
        
        if (salt == null) {
            saltedpass =  password.getBytes("UTF-8");
        }
        else {
            byte[] pass = password.getBytes("UTF-8");
            saltedpass = new byte[pass.length+salt.length];
            
            System.arraycopy(salt, 0, saltedpass, 0, salt.length);
            System.arraycopy(pass, 0, saltedpass, salt.length, pass.length);
        }
                
        // Create key by hashing the password (+ the salt)
        byte[] key = new byte[keylength/8];

        MessageDigest msgDigest = MessageDigest.getInstance("SHA-256");
        msgDigest.reset();

        byte[] longkey = msgDigest.digest(saltedpass);

        if (longkey.length == key.length) {
            key = longkey;
        }

        else if (longkey.length > key.length) {
            System.arraycopy(longkey, 0, key, 0, key.length);
        }

        else if (longkey.length < key.length) {
            throw new RuntimeException("Invalid key length '"+keylength+"' bit; max 256 bit supported.");
        }

        System.out.println("key = "+Arrays.toString(key));
        return new SecretKeySpec(key, cipherStr); // AES -> 128/192 bit        
    }    

    @Override
    public void load(ConfigNode node) throws ConfigException {
        password = node.getProperty("password");
        cipherStr = node.getProperty("cipher");
        keylength = node.getInteger("keylength");

        if (password == null)
            throw new ConfigException("Password must be specified in encryption.");

        if (cipherStr == null)
            cipherStr = Constants.DEFAULT_ENCRYPTION_CIPHER;

        if (keylength == null)
            keylength = Constants.DEFAULT_ENCRYPTION_KEYLENGTH;

        init();
    }

    @Override
    public void save(ConfigNode node) {
        node.setProperty("password", password);
        node.setProperty("cipher", cipherStr);
        node.setProperty("keylength", keylength);
    }    
    
    public static void main(String[] args) throws ConfigException, IOException, EncryptionException {
        Config.getInstance().load();
        Cipher encCipher = Config.getInstance().getProfiles().get(1).getRepository().getEncryption()
                .createEncCipher("somesalt");
        
        Cipher decCipher = Config.getInstance().getProfiles().get(1).getRepository().getEncryption()
                .createDecCipher("somesalt2");

        ByteArrayInputStream bIn = new ByteArrayInputStream("HALLO TEST".getBytes());
        CipherInputStream cIn = new CipherInputStream(bIn, encCipher);
        ByteArrayOutputStream bOut = new ByteArrayOutputStream();

        int ch;
        while ((ch = cIn.read()) >= 0) {
          bOut.write(ch);
        }

        byte[] cipherText = bOut.toByteArray();

        System.out.println("cipher: " + new String(cipherText));

        // decryption pass
        bOut = new ByteArrayOutputStream();
        CipherOutputStream cOut = new CipherOutputStream(bOut, decCipher);
        cOut.write(cipherText);
        cOut.close();
        System.out.println("plain : " + new String(bOut.toByteArray()));        
    }    
}
