/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.connection.plugins.dropbox;

import java.util.HashMap;
import java.util.Map;
import org.syncany.config.ConfigNode;
import org.syncany.connection.plugins.ConfigPanel;
import org.syncany.connection.plugins.Connection;
import org.syncany.connection.plugins.PluginInfo;
import org.syncany.connection.plugins.Plugins;
import org.syncany.connection.plugins.TransferManager;
import org.syncany.exceptions.ConfigException;

/**
 *
 *
 * @author Philipp C. Heckel <philipp.heckel@gmail.com>
 */
public class DropboxConnection implements Connection {    
    private static final Map CONSTANTS = new HashMap();
    
    private String path;

    // For the consumer key/secret, add new app to Dropbox here:
    // https://www.dropbox.com/developers/apps
    private String consumerKey;
    private String consumerSecret;

    // To acquire an access key/secret, a user must login to the
    // application. See function connect()
    private String accessTokenKey;
    private String accessTokenSecret;
    
    
    static {
        // From https://www.dropbox.com/developers/releases
        // config/testing.json.example
        CONSTANTS.put("server", "api.dropbox.com");
        CONSTANTS.put("content_server", "api-content.dropbox.com");
        CONSTANTS.put("port", 80);
        CONSTANTS.put("request_token_url", "http://api.dropbox.com/0/oauth/request_token");
        CONSTANTS.put("access_token_url", "http://api.dropbox.com/0/oauth/access_token");
        CONSTANTS.put("authorization_url", "http://api.dropbox.com/0/oauth/authorize");     
        
        // Must be set by the user:
        //   consumer_key + consumer_secret
    }    

    @Override
    public PluginInfo getPluginInfo() {
        return Plugins.get(DropboxPluginInfo.ID);
    }
 
    @Override
    public TransferManager createTransferManager() {
        return new DropboxTransferManager(this);
    }

    @Override
    public ConfigPanel createConfigPanel() {
        return new DropboxConfigPanel(this);
    }

    public String getConsumerKey() {
        return consumerKey;
    }

    public void setConsumerKey(String consumerKey) {
        this.consumerKey = consumerKey;
    }

    public String getConsumerSecret() {
        return consumerSecret;
    }

    public void setConsumerSecret(String consumerSecret) {
        this.consumerSecret = consumerSecret;
    }

    public String getAccessTokenKey() {
        return accessTokenKey;
    }

    public void setAccessTokenKey(String accessTokenKey) {
        this.accessTokenKey = accessTokenKey;
    }

    public String getAccessTokenSecret() {
        return accessTokenSecret;
    }

    public void setAccessTokenSecret(String accessTokenSecret) {
        this.accessTokenSecret = accessTokenSecret;
    }

    public String getPath() {
        return path;
    }

    public void setPath(String path) {
        if (path.endsWith("/")) {
            this.path = path.substring(0, path.length()-1);
        }
        else {
            this.path = path;
        }
    }       

    public Map getConfig() {
        Map config = new HashMap(CONSTANTS);
        
        config.put("consumer_key", consumerKey);
        config.put("consumer_secret", consumerSecret);        
        config.put("access_token_key", accessTokenKey);
        config.put("access_token_secret", accessTokenSecret);        
        
        return config;
    }    

    @Override
    public void load(ConfigNode node) throws ConfigException {
        // Mandatory
        path = node.getProperty("path");
        consumerKey = node.getProperty("consumerkey");
        consumerSecret = node.getProperty("consumersecret");                

        if (consumerKey == null || consumerSecret == null || path == null) {
            throw new ConfigException("Dropbox connection properties must at least contain the parameters 'consumerkey', 'consumersecret' and 'path'.");
        }
        
        // Optional (manual user login required if not present)
        accessTokenKey = node.getProperty("accesstokenkey");
        accessTokenSecret = node.getProperty("accesstokensecret");                
    }

    @Override
    public void save(ConfigNode node) {
        node.setAttribute("type", DropboxPluginInfo.ID);

        node.setProperty("path", path);
        node.setProperty("consumerkey", consumerKey);
        node.setProperty("consumersecret", consumerSecret);
        node.setProperty("accesstokenkey", accessTokenKey);
        node.setProperty("accesstokensecret", accessTokenSecret);
    }
    
    @Override
    public String toString() {
        return DropboxConnection.class.getSimpleName()
        + "[consumer_key=" + consumerKey + ", path="+path+"]";
    }	
}
