/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.connection.plugins.dropbox;

import com.dropbox.client.Authenticator;
import com.dropbox.client.DropboxClient;
import org.apache.http.HttpResponse;
import org.syncany.connection.plugins.AbstractTransferManager;
import org.syncany.exceptions.StorageConnectException;
import org.syncany.watch.remote.files.RemoteFile;
import org.syncany.exceptions.StorageException;
import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import oauth.signpost.exception.OAuthException;
import org.json.simple.JSONArray;
import org.json.simple.JSONObject;
import org.syncany.exceptions.RemoteFileNotFoundException;
import org.syncany.util.FileUtil;

/**
 *
 * 
 * @author Philipp C. Heckel <philipp.heckel@gmail.com>
 */
public class DropboxTransferManager extends AbstractTransferManager { 
    private static final String DROPBOX_ROOT = "dropbox"; // alternatively: 'sandbox'
    private static final int DROPBOX_MAX_LIST = 10000; // max. as of API spec.
    
    private static final Logger logger = Logger.getLogger(DropboxTransferManager.class.getSimpleName());
    
    private Authenticator auth;
    private DropboxClient dropbox;
    
    private boolean connected;
    private boolean userLoginRequired;
    private String userLoginURL;
    
    public DropboxTransferManager(DropboxConnection connection) {
        super(connection);
    } 
 
    @Override
    public DropboxConnection getConnection() {
        return (DropboxConnection) super.getConnection();
    }

    @Override
    public void connect() throws StorageConnectException {
        if (connected) {
            return;
        }
        
        // Not connected / tokens verified
        try {
            if (auth == null) {
                auth = new Authenticator(getConnection().getConfig());                
            }                                          

            // a) We do not have an access token key in the config
            //    A manual user login is required!
            if (getConnection().getAccessTokenKey() == null) {
                try {
                    if (!userLoginRequired) {
                        String url = auth.retrieveRequestToken(""); 
                        userLoginURL = url.replace("api.dropbox.com", "www.dropbox.com");                    
                    }

                    // Test token for validity!                
                    auth.retrieveAccessToken("");
                    
                    // Set access token
                    dropbox = new DropboxClient(getConnection().getConfig(), auth);                        
                    
                    getConnection().setAccessTokenKey(auth.getTokenKey());
                    getConnection().setAccessTokenSecret(auth.getTokenSecret());
                }
                catch (OAuthException e) {
                    if (logger.isLoggable(Level.WARNING)) {
                        logger.warning("No access token in config; User login required; Acquiring request token ...");
                    }

                    userLoginRequired = true;
                    throw new StorageConnectException("User login required at "+userLoginURL);
                }
            }
            
            // b) We do have an access token. Test if it works!
            else {
                try {
                    dropbox = new DropboxClient(getConnection().getConfig(), auth);                                            
                    Map response = dropbox.accountInfo(false, null);
                    //for (Object k : response.keySet()) System.out.println(k+" = "+response.get(k));
                    
                    if (response == null || !response.containsKey("uid")) {
                        throw new Exception("Invalid access token credentials.");
                    }
                }
                catch (Exception e) {
                    if (logger.isLoggable(Level.WARNING)) {
                        logger.warning("Access token found in config, but it seems to be invalid. Re-login required.");                        
                    }
                    
                    // Invalidate access token
                    getConnection().setAccessTokenKey(null);
                    getConnection().setAccessTokenSecret(null);
                    
                    // NOTE: Recursive call; Now without access token!!
                    connect();                    
                    return;
                }
            }

            // We did it! We are logged in!
            userLoginRequired = false;

            // Test given folder
            Map response = dropbox.metadata(DROPBOX_ROOT, getConnection().getPath(), 1, "", false, false, null);
            // for (Object k : response.keySet()) System.out.println(k+" = "+response.get(k));
            
            if (!response.containsKey("is_dir") || !"true".equals(response.get("is_dir").toString())) {
                throw new StorageConnectException("Invalid repository path. Not existant or not a directory? "+getConnection().getPath());
            }
            
            connected = true;
        }
        catch (Exception e) {
            if (logger.isLoggable(Level.WARNING)) {
                logger.log(Level.WARNING, "Dropbox client connection failed.", e);
            }
            
            throw new StorageConnectException(e);
        }        
    }

    public boolean isUserLoginRequired() {
        return userLoginRequired;
    }

    public String getUserLoginURL() {
        return userLoginURL;
    }
    
    @Override
    public void disconnect() {
        // Nothing to do.
    }

    @Override
    public void download(RemoteFile remoteFile, File localFile) throws StorageException {
        connect();
        String remotePath = getConnection().getPath()+"/"+remoteFile.getName();

        try {
            // Download file
            File tempFile = config.getCache().createTempFile();
            OutputStream tempFOS = new FileOutputStream(tempFile);

            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "Dropbox: Downloading {0} to temp file {1}", new Object[]{remotePath, tempFile});            
            }
                        
            HttpResponse response = dropbox.getFile(DROPBOX_ROOT, remotePath);
            
            if (response == null || response.getStatusLine().getStatusCode() != 200) {
                throw new RemoteFileNotFoundException("Remote file not found: "+remotePath);
            }
            
            FileUtil.copy(response.getEntity().getContent(), tempFOS);
            tempFOS.close();

            // Move file
            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "Dropbox: Renaming temp file {0} to file {1}", new Object[]{tempFile, localFile});            
            }            
            
            if (!tempFile.renameTo(localFile)) {
                throw new StorageException("Rename to "+localFile.getAbsolutePath()+" failed.");
            }
        }
        catch (Exception ex) {
            if (logger.isLoggable(Level.SEVERE)) {
                logger.log(Level.SEVERE, "Error while downloading file "+remoteFile.getName(), ex);
            }
            
            throw new StorageException(ex);
        }
    }

    @Override
    public void upload(File localFile, RemoteFile remoteFile) throws StorageException {
        connect();

        try {
            // Upload to temp file
            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "Dropbox: Uploading {0} to file {1}", new Object[]{localFile, remoteFile.getName()});            
            }
            
            HttpResponse response = dropbox.putFile(DROPBOX_ROOT, getConnection().getPath(), localFile, remoteFile.getName());
            
            if (response == null || response.getStatusLine().getStatusCode() != 200) {
                throw new RemoteFileNotFoundException("Could not upload file to file: "+remoteFile.getName());
            }
        }
        catch (Exception ex) {
            logger.log(Level.SEVERE, "Could not upload file "+localFile+" to "+remoteFile.getName(), ex);
            throw new StorageException(ex);
        }
    }

    @Override
    public Map<String, RemoteFile> list() throws StorageException {
        connect();

        try {
            Map<String, RemoteFile> files = new HashMap<String, RemoteFile>();
            
            Map folder = dropbox.metadata(DROPBOX_ROOT, getConnection().getPath(), DROPBOX_MAX_LIST, "", true, false, null);
            for (Object k : folder.keySet()) System.out.println(k +" = "+folder.get(k));
            
            JSONArray dropboxFiles = (JSONArray) folder.get("contents");
            
            for (Object o : dropboxFiles) {
                JSONObject jo = (JSONObject) o;
                
                // Returns "/root/filename"; now strip to "filename"
                String filename = (String) jo.get("path");
                String basename = filename.substring(getConnection().getPath().length()+1);
                
                // Size
                long filesize = (Long) jo.get("bytes");
                
                //System.out.println(k +" = "+dropboxFiles.get(k));
                files.put(basename, new RemoteFile(basename, filesize, jo));
            }

            return files;
        }
        catch (Exception ex) {
            logger.log(Level.SEVERE, "Unable to list FTP directory.", ex);
            throw new StorageException(ex);
        }
    }

    @Override
    public void delete(RemoteFile remoteFile) throws StorageException {
        connect();

        try {
            dropbox.fileDelete(DROPBOX_ROOT, getConnection().getPath()+"/"+remoteFile.getName(), null);
        }
        catch (Exception ex) {
            logger.log(Level.SEVERE, "Could not delete file "+remoteFile.getName(), ex);
            throw new StorageException(ex);
        }
    }
}
