/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.db;

import java.io.Serializable;
import java.util.Arrays;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.PostPersist;
import org.apache.commons.codec.binary.Base32;
import org.eclipse.persistence.annotations.Index;
import org.syncany.util.StringUtil;

/**
 * Represents the chunk of a file.
 * 
 * @author Philipp C. Heckel
 */
@Entity
public class CloneChunk extends PersistentObject implements Serializable {
    private static final Base32 b32 = new Base32();        
    
    @Id
    @GeneratedValue(strategy=GenerationType.SEQUENCE)
    @Column(name="chunk_dbid")
    private long dbId;
    
    @Index
    @Column(name="metaid", columnDefinition="binary(20)")
    private byte[] metaId;
    
    @Index
    @Column(name="checksum", columnDefinition="binary(20) not null unique")
    private byte[] checksum;

    public CloneChunk() {
        // Nothing.
    }
    
    public CloneChunk(byte[] checksum) {
        this();
        this.checksum = checksum;
    }
    
    @PostPersist
    public void postPersist() {
        Database.getInstance().addToMemCache(this);
    }    

    public byte[] getMetaId() {
        return metaId;
    }

    public void setMetaId(byte[] metaId) {
        this.metaId = metaId;
    }

    public byte[] getChecksum() {
        return checksum;
    }
    
    public void setChecksum(byte[] checksum) {
        this.checksum = checksum;
    }
    
    public String getIdStr()  {
        return encodeIdStr(metaId, checksum);
    }

    public String getFileName() {
        return getFileName(getMetaId(), getChecksum());
    }
    
    public static String getFileName(String idStr) {
        byte[] checksum = CloneChunk.decodeChecksum(idStr);
        byte[] metaId = CloneChunk.decodeMetaId(idStr);
        
        return getFileName(metaId, checksum);
    }    
    
    public static String getFileName(byte[] metaId, byte[] checksum) {
        return String.format("chunk-%s", encodeIdStr(metaId, checksum));
    }
    
    public static String encodeIdStr(byte[] checksum) {
        return encodeIdStr(null, checksum);
    }
    
    public static String encodeIdStr(byte[] metaId, byte[] checksum) {
        if (metaId == null) {
            if (checksum == null) {
                return null;
            }
            
            //return String.format("%s", b32.encodeAsString(checksum).replace("=", "")).toLowerCase();
            return String.format("%s", StringUtil.toHex(checksum)).toLowerCase();
        }
        else if (checksum == null) {
            //return String.format("%s", b32.encodeAsString(metaId).replace("=", "")).toLowerCase();
            return String.format("%s", StringUtil.toHex(metaId)).toLowerCase();
        }
        else {
            return String.format("%s-%s", 
                    StringUtil.toHex(metaId), 
                    StringUtil.toHex(checksum)).toLowerCase();
            //return String.format("%s-%s", 
            //        b32.encodeAsString(metaId).replace("=", ""), 
            //        b32.encodeAsString(checksum).replace("=", "")).toLowerCase();
        }       
    }
    
    public static byte[] decodeChecksum(String idStr) {
        int dashpos = idStr.indexOf("-");
        return (dashpos == -1) ? StringUtil.fromHex(idStr.toUpperCase()) : StringUtil.fromHex(idStr.toUpperCase().substring(dashpos+1));
        //return (dashpos == -1) ?  b32.decode(idStr.toUpperCase()) : b32.decode(idStr.toUpperCase().substring(dashpos+1));
    }
    
    public static byte[] decodeMetaId(String idStr) {
        int dashpos = idStr.indexOf("-");
        return (dashpos == -1) ? null : StringUtil.fromHex(idStr.toUpperCase().substring(0, dashpos));
        //return (dashpos == -1) ? null : b32.decode(idStr.toUpperCase().substring(0, dashpos));
    }    
    /*public static void main(String[] args) {
        String idStr = encodeIdStr(
                new byte[]{9,8,7,6,5,4,3}
                , new byte[] { 0,1,2,3,4,5,6});//, );
        System.out.println(idStr);
        System.out.println(Arrays.toString(decodeChecksum(idStr)));
        System.out.println(Arrays.toString(decodeMetaId(idStr)));
        
    }*/
    public static String getMetaIdStr(String idStr) {
        int dashpos = idStr.indexOf("-");
        return (dashpos == -1) ? null : idStr.substring(0, dashpos);
    }
    
    public static String getChecksumStr(String idStr) {
        int dashpos = idStr.indexOf("-");
        return (dashpos == -1) ? idStr : idStr.substring(dashpos+1);
    }


    @Override
    public int hashCode() {
        int hash = 5;
        hash = 79 * hash + Arrays.hashCode(this.checksum);
        return hash;
    }  

    @Override
    public boolean equals(Object object) {
        if (!(object instanceof CloneChunk)) {
            return false;
        }
        
        CloneChunk other = (CloneChunk) object;

        if (this.checksum != other.checksum) {
            return false;
        }
        
        return true;
    }

    @Override
    public String toString() {
        return "CloneChunk[dbId="+dbId+", metaId="+encodeIdStr(metaId)+", checksum="+encodeIdStr(checksum) + "]";
    }
}
