/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.db;

import java.io.IOException;
import org.syncany.config.Config;
import org.syncany.config.Folder;
import org.syncany.config.Profile;
import org.syncany.exceptions.ConfigException;
import org.syncany.util.FileUtil;
import java.io.File;
import java.io.Serializable;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.Random;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.persistence.CascadeType;
import javax.persistence.Column;
import javax.persistence.Entity;
import javax.persistence.EnumType;
import javax.persistence.Enumerated;
import javax.persistence.GeneratedValue;
import javax.persistence.GenerationType;
import javax.persistence.Id;
import javax.persistence.IdClass;
import javax.persistence.JoinColumn;
import javax.persistence.JoinColumns;
import javax.persistence.ManyToMany;
import javax.persistence.ManyToOne;
import javax.persistence.OneToMany;
import javax.persistence.OneToOne;
import javax.persistence.OrderBy;
import javax.persistence.OrderColumn;
import javax.persistence.Query;
import javax.persistence.Temporal;
import javax.persistence.TemporalType;
import javax.persistence.Transient;
import org.apache.commons.codec.binary.Base32;
import org.bouncycastle.util.Arrays;
import org.eclipse.persistence.annotations.Index;
import org.syncany.config.Repository;
import org.syncany.watch.remote.files.UpdateFile;

/**
 * Represents a version of a file. 
 * 
 * @author Philipp C. Heckel
 */
@Entity
//@Table( uniqueConstraints={@UniqueConstraint(columnNames={"last_modified", "checksum"})} )
public class CloneFile extends PersistentObject implements Serializable, Cloneable {
    private static final Logger logger = Logger.getLogger(CloneFile.class.getSimpleName());
    private static final Config config = Config.getInstance();
    private static final Base32 b32 = new Base32();
    /**
     * <ul>
     * <li>UNKNOWN
     * <li>NEW: New file
     * <lI>CHANGED: The file contents have changed. At least one chunk differs.
     * <li>RENAMED: The file path or name has changed.
     * <li>MERGED: The file history has been merged to a different file.
     * </ul>
     */
    public enum Status { UNKNOWN, NEW, CHANGED, RENAMED, DELETED, MERGED };
    
    /**
     * LOCAL: The file entry hasn't been propagated to the server yet
     * IN_UPDATE: The file entry should be included in the update-file, but not (yet) in the base file
     * IN_BASE: The file entry should be included in the base-file (= complete DB dump)
     */
    public enum SyncStatus { UNKNOWN, LOCAL, SYNCING, UPTODATE, CONFLICT };

    /*@Id
    @GeneratedValue(strategy=GenerationType.SEQUENCE)
    @Column(name="file_dbid")
    private long dbId;*/
    
    /**
     * versionId of the root file; identifies the history of a file
     */
    @Id
    @Column(name="file_id", nullable=false)
    private Long fileId;

    @Id
    @Column(name="file_version", nullable=false)
    private long version;

    @Id
    @Column(name="profile_id", nullable=false)
    private int profileId;

    @Id
    @Column(name="root_id", nullable=false)
    private String rootId;

    @Temporal(TemporalType.TIMESTAMP)
    @Column(name="updated")
    private Date updated;
    
    @Column(name="checksum", columnDefinition="binary(20)")
    private byte[] checksum;

    @Transient
    private Profile profile;

    @Transient
    private Folder root;
   


    // FILE PROPERTIES

    @Column(name="is_folder")
    private boolean folder;
    
    @ManyToOne
    @JoinColumns({
        @JoinColumn(name="merged_to_file_id", referencedColumnName="file_id"),
        @JoinColumn(name="merged_to_file_version", referencedColumnName="file_version"),
        @JoinColumn(name="merged_to_profile_id", referencedColumnName="profile_id"),
        @JoinColumn(name="merged_to_root_id", referencedColumnName="root_id")
    })
    private CloneFile mergedTo;    

    @Column(name="path", nullable=false)
    private String path;

    @Column(name="name", nullable=false)
    private String name;

    @Column(name="file_size")
    private long fileSize;
    
    @Temporal(TemporalType.TIMESTAMP)
    @Column(name="last_modified")
    private Date lastModified;
    
    // JPA does not support duplicates in a list. This is a workaround.
    // http://en.wikibooks.org/wiki/Java_Persistence/Relationships#Duplicates
    
    @OneToMany(cascade=CascadeType.ALL)
    @OrderBy(value="orderNumber")
    @JoinColumns({
        @JoinColumn(name="file_id", referencedColumnName="file_id"),
        @JoinColumn(name="file_version", referencedColumnName="file_version"),
        @JoinColumn(name="profile_id", referencedColumnName="profile_id"),
        @JoinColumn(name="root_id", referencedColumnName="root_id")
    })
    private List<ChunkMapper> chunks;

    @Enumerated(EnumType.STRING)
    @Column(name="status")
    private Status status;

    @Enumerated(EnumType.STRING)
    @Column(name="sync_status")
    private SyncStatus syncStatus;

    // TODO this should be a CloneClient instance
    @Column(name="client_name")
    private String clientName;
    
    public static void main(String[] args) throws ConfigException, InterruptedException, IOException {
        Config config = Config.getInstance();
        config.load();
        
        
        CloneChunk chunk = Database.getInstance().getChunk(new byte[]{5,5,5}, true);
        chunk = Database.getInstance().getChunk(new byte[]{5,5,5}, true);
        
        /*

        CloneChunk chunk2 = new CloneChunk(new byte[]{1,2,3,4});
        chunk2.merge();
        */
        CloneFile f = new CloneFile();
        f.setPath("");
        f.setName("bbb");
        f.setFileId(1L);
        f.setVersion(1L);
        f.setProfileId(1);
        f.setRootId("abc");
        f.setChecksum(new byte[] {1,2,3});
        f.addChunk(chunk);
        //f.addChunk(chunk2);
        f.addChunk(chunk);
        //f.addChunk(chunk2);
        
        f.merge();
        
        
        f = new CloneFile();
        f.setPath("");
        f.setName("NEWFILE 2");
        f.setFileId(2L);
        f.setVersion(1L);
        f.setProfileId(1);
        f.setRootId("abc");
        f.setChecksum(new byte[] {9,9,9});
        f.addChunk(chunk);
        //f.addChunk(chunk2);
        f.addChunk(chunk);        
        
        f.merge();
        
        System.out.println(f.getChunks());
        
        //System.out.println(Database.getInstance().getChunk(new byte[] {1,2,3}, true));
        //System.out.println(Database.getInstance().getChunk(new byte[] {1,2,3}, true));
        UpdateFile uf = new UpdateFile(new Repository(), "platop", new Date());
        ArrayList<CloneFile> a = new ArrayList<CloneFile>();
        a.add(f);
        
        uf.setVersions(a);
        //uf.write(new File("/home/pheckel/tmp"), false);
    }
    
    /*public Object doMerge() {
        config.getDatabase().getEntityManager().getTransaction().begin();

        for (int i = 0; i < chunks.size(); i++) {
            ChunkMapper chunkMapper = chunks.get(i);
            CloneChunk chunk = chunkMapper.getChunk();
            
            CloneChunk mergedChunk = config.getDatabase().getEntityManager().merge(chunk);
            chunkMapper.setChunk(mergedChunk);            

            ChunkMapper mergedChunkMapper = config.getDatabase().getEntityManager().merge(chunkMapper);
            
            chunks.set(i, mergedChunkMapper);
        }
        
        config.getDatabase().getEntityManager().merge(this);
        
        config.getDatabase().getEntityManager().flush();
        config.getDatabase().getEntityManager().getTransaction().commit();
    }*/
    
    
    
    public CloneFile() {
        this.fileId = new Random().nextLong();
        this.version = 1;
        /*this.chunkMapping = new ArrayList<byte[]>();
        this.chunks = new ArrayList<CloneChunk>();*/
        this.chunks = new ArrayList<ChunkMapper>();
        this.status = Status.UNKNOWN;
        this.syncStatus = SyncStatus.UNKNOWN;

        this.checksum = null;
        this.name = "(unknown)";
        this.path = "(unknown)";
        this.clientName = config.getMachineName();
    }

    public CloneFile(Folder root, File file) {
        this();

        // Set account
        setProfile(root.getProfile());
        setRoot(root);

        setName(file.getName());
        setPath(FileUtil.getRelativeParentDirectory(root.getLocalFile(), file));
        setFileSize(file.isDirectory() ? 0 : file.length());
        setLastModified(new Date(file.lastModified()));
        setFolder(file.isDirectory());
    }

    public Folder getRoot() {
        if (root == null) {
            root = getProfile().getFolders().get(getRootId());
        }

        return root;
    }

    public void setRoot(Folder root) {
        this.root = root;
        this.rootId = root.getRemoteId();
    }

    public CloneFile getMergedTo() {
        return mergedTo;
    }

    public void setMergedTo(CloneFile mergedTo) {
        this.mergedTo = mergedTo;
    }
    
    public Profile getProfile() {
        if (profile == null) {
            profile = config.getProfiles().get(getProfileId());
        }

        return profile;
    }

    public void setProfile(Profile profile) {
        this.profile = profile;
        this.profileId = profile.getId();
    }

    public int getProfileId() {
        return profileId;
    }

    public void setProfileId(int profileId) {
        this.profileId = profileId;
    }

    public String getRootId() {
        return rootId;
    }

    public void setRootId(String rootId) {
        this.rootId = rootId;
    }

    public boolean isFolder() {
        return folder;
    }

    public void setFolder(boolean folder) {
        this.folder = folder;
    }

    public long getVersion() {
        return version;
    }

    public void setVersion(long version) {
        this.version = version;
    }

    public Long getFileId() {
        return fileId;
    }

    public void setFileId(Long fileId) {
        this.fileId = fileId;
    }

    public Date getUpdated() {
        return updated;
    }

    public void setUpdated(Date updated) {
        this.updated = updated;
    }

    public String getPath() {        
        return Database.toFilesystemPath(path);
    }

    public void setPath(String path) {
        this.path = Database.toDatabasePath(path);
    }

    public SyncStatus getSyncStatus() {
        return syncStatus;
    }

    public void setSyncStatus(SyncStatus syncStatus) {
        this.syncStatus = syncStatus;
    }

    public Date getLastModified() {
        return lastModified;
    }

    public void setLastModified(Date lastModified) {
        this.lastModified = lastModified;
    }

    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }
   
    public static String encodeChecksum(byte[] checksum) {
        return String.format("%s", b32.encodeAsString(checksum).replace("=", "")).toLowerCase();
    }
    
    public static byte[] decodeChecksum(String checksumStr) {
        return b32.decode(checksumStr.toUpperCase());
    }    
    /**
     * Get relative path to the root dir.
     */
    public String getRelativePath() {
        return FileUtil.getRelativePath(getRoot().getLocalFile(), getFile());
    }

    public String getRelativeParentDirectory() {
        return FileUtil.getRelativeParentDirectory(getRoot().getLocalFile(), getFile());
    }

    public String getAbsoluteParentDirectory() {
        return FileUtil.getAbsoluteParentDirectory(getFile());
    }

    public File getFile() {
        return FileUtil.getCanonicalFile(
            new File(getRoot().getLocalFile()+File.separator+getPath()+File.separator+getName()));
    }

    public long getFileSize() {
        return fileSize;
    }

    public void setFileSize(long fileSize) {
        this.fileSize = fileSize;
    }

    public byte[] getChecksum() {
        return checksum;
    }

    public String getChecksumStr() {
        return (checksum == null) ? "" : encodeChecksum(checksum);
    }
    
    public void setChecksum(byte[] checksum) {
        this.checksum = checksum;
    }

    public String getClientName() {
        return clientName;
    }

    public void setClientName(String clientName) {
        this.clientName = clientName;
    }        

    public CloneFile getPreviousVersion() {
        // If we are the first, there are no others
        if (getVersion() == 1) {
            return null;
        }

        List<CloneFile> pV = getPreviousVersions();

        if (pV.isEmpty()) {
            return null;
        }

        return pV.get(pV.size()-1);
    }

    public List<CloneFile> getPreviousVersions() {
        // If we are the first, there are no others
        if (getVersion() == 1) {
            return new ArrayList<CloneFile>();
        }

        Query q = config.getDatabase().getEntityManager().createQuery(
              "select c from CloneFile c "
            + "where c.profileId = :profileId and c.fileId = :fileId and c.version < :version "
            + "order by c.version asc"
        );

        q.setParameter("profileId", getProfileId());
        q.setParameter("fileId", getFileId());
        q.setParameter("version", getVersion());

        return q.getResultList();
    }

    public List<CloneFile> getNextVersions() {
        Query q = config.getDatabase().getEntityManager().createQuery(
              "select c from CloneFile c "
            + "where c.profileId = :profileId and c.fileId = :fileId and c.version > :version "
            + "order by c.version asc"
        );

        q.setParameter("profileId", getProfileId());
        q.setParameter("fileId", getFileId());
        q.setParameter("version", getVersion());

        return q.getResultList();
    }

    public List<CloneFile> getVersionHistory() {
        List<CloneFile> versions = new ArrayList<CloneFile>();

        versions.addAll(getPreviousVersions());
        versions.add(this);
        versions.addAll(getNextVersions());

        return versions;
    }    
    
    public Status getStatus() {
        return status;
    }

    public void setStatus(Status status) {
        this.status = status;
    }

    public List<CloneChunk> getChunks() {
        List<CloneChunk> chunkList = new ArrayList<CloneChunk>();
        
        for (ChunkMapper mapper : chunks) {
            chunkList.add(mapper.getChunk());
        }
        
        return chunkList;
    }

    public CloneChunk getChunk(int index) {
        return chunks.get(index).getChunk();
    }

    public synchronized void setChunks(List<CloneChunk> chunkList) {
        chunks.clear();
        
        for (int i = 0; i < chunkList.size(); i++) {
            chunks.add(new ChunkMapper(chunkList.get(i), i));
        }
    }

    public synchronized void setChunk(int index, CloneChunk chunk) {
        chunks.set(index, new ChunkMapper(chunk, index));
    }

    public void addChunk(CloneChunk chunk) {
        chunks.add(new ChunkMapper(chunk, chunks.size()));
    }

    public void addChunks(List<CloneChunk> chunkList) {
        for(CloneChunk chunk : chunkList) {
            addChunk(chunk);
        }
    }

    public synchronized void removeChunks(int count) {
        int minIndex = (chunks.size()-count-1 < 0) ? 0 : chunks.size()-count-1;
        int maxIndex = (chunks.size()-1 < 0) ? 0 : chunks.size()-1;

        for (int i=maxIndex; i==minIndex; i--) {
            chunks.remove(i);
        }
    }
  
    @Override
    public int hashCode() {
        int hash = 0;
        hash += (fileId != null ? fileId.hashCode() : 0);
        hash += version;
        return hash;
    }
    
    @Override
    public Object clone() {
        try {
            CloneFile clone = (CloneFile) super.clone();

            clone.fileId = getFileId();
            clone.updated = getUpdated();
            clone.checksum = Arrays.clone(getChecksum());
            clone.lastModified = new Date(getLastModified().getTime());
            clone.profile = getProfile(); // POINTER; No Copy!
            clone.profileId = getProfileId();
            clone.root = getRoot(); // POINTER; No Copy!
            clone.rootId = getRootId();
            clone.folder = isFolder();            
            clone.path = new String(getPath()); // Copy
            clone.name = new String(getName()); // Copy
            clone.clientName = new String(getClientName()); // Copy
            clone.fileSize = getFileSize();
            clone.status = getStatus(); 
            clone.syncStatus = getSyncStatus();
            clone.mergedTo = getMergedTo(); // POINTER

            // Copies
            clone.setChunks(getChunks()); // Copy (due to mapper)

            return clone;
        }
        catch (Exception ex) {
            logger.log(Level.SEVERE, null, ex);
            throw new RuntimeException(ex);
        }
    }    

    @Override
    public boolean equals(Object object) {
        if (!(object instanceof CloneFile)) {
            return false;
        }
        
        CloneFile other = (CloneFile) object;

        if (other.fileId == null || this.fileId == null) {
            return false;
        }

        return other.fileId.equals(this.fileId) && other.version == this.version;
    }

    @Override
    public String toString() {
        return "CloneFile[id=" + fileId + ",version="+version+", path="+getPath()+", name="+getName()+"]";
    }

}
