/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.index.chunk;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.security.MessageDigest;
import java.util.Enumeration;
import java.util.logging.Level;

/**
 *
 * @author Philipp C. Heckel <philipp.heckel@gmail.com>
 */
public class TTTDChunker extends Chunker {   
    public static final int DEFAULT_WINDOW_SIZE = 48; // like LBFS
    public static final String DEFAULT_DIGEST_ALG = "SHA1";
    
    private int Tmin;
    private int Tmax;
    private int D;
    private int Ddash;   
    private int windowSize;
    private String digestAlg;
    
    public TTTDChunker(int Tmin, int Tmax, int D, int Ddash) {
        this(Tmin, Tmax, D, Ddash, DEFAULT_WINDOW_SIZE, DEFAULT_DIGEST_ALG);
    }
    
    public TTTDChunker(int Tmin, int Tmax, int D, int Ddash, int windowSize, String digestAlg) {
        this.Tmin = Tmin;
        this.Tmax = Tmax;
        this.D = D;
        this.Ddash = Ddash;
        this.windowSize = windowSize;
        this.digestAlg = digestAlg;
        
        if (windowSize > Tmin) {
            throw new IllegalArgumentException("Window size must be smaller than Tmin.");
        }           
    }        
   
    @Override
    public Enumeration<Chunk> createChunks(File file) throws FileNotFoundException {
        return new TTTDEnumeration(file);
    } 
    
    public class TTTDEnumeration implements Enumeration<Chunk> {        
        private InputStream fis;           
        private boolean closed;
        private byte[] c;
        private int clen;
        private int cpos;
        
        private MessageDigest chunkDigest;
        private MessageDigest fileDigest;
        private RollingChecksum breakCheck;

        public TTTDEnumeration(File file) throws FileNotFoundException {
            this.fis = new FileInputStream(file);
            this.closed = false;
            this.c = new byte[8192];
            this.clen = -1;
            this.cpos = -1;
            
            this.breakCheck = new RollingChecksum();    

            try {
                this.chunkDigest = MessageDigest.getInstance(digestAlg);

                this.fileDigest = MessageDigest.getInstance(digestAlg);
                this.fileDigest.reset();
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }                          
        }
        
        @Override
        public boolean hasMoreElements() {
            return !closed;
        }

        @Override
        public Chunk nextElement() {
            if (closed) {
                return null;
            }
            
            chunkDigest.reset();
            breakCheck.reset();
                
            try {
                int backupBreak = 0;                
                int breakpoint = -1;
                                
                byte[] buf = new byte[Tmax];
                int bufpos = -1;
                                
                while (bufpos < buf.length-1) {
                    if (cpos == -1 || cpos == clen-1) {
                        cpos = -1;
                        clen = fis.read(c);
                        
                        if (clen == -1) {
                            break;
                        }
                                            
                        fileDigest.update(c, 0, clen);                        
                    }
                        
                    bufpos++; cpos++;
                    buf[bufpos] = c[cpos];   

                    if (bufpos < Tmin) {
                        continue;
                    }
                    else if (bufpos == Tmin) {
                        breakCheck.check(buf, bufpos-windowSize, windowSize);
                    }
                    else {
                        breakCheck.roll(buf[bufpos]);
                    }

                    int hash = breakCheck.getValue();

                    if ((hash % Ddash) == Ddash-1) {      
                        backupBreak = bufpos;     
                    }

                    if ((hash % D) == D-1) {
                        breakpoint = bufpos;
                        break;
                    }

                    if (bufpos < Tmax){
                        continue;
                    }

                    if (backupBreak != 0) {
                        breakpoint = backupBreak;
                        break;
                    }
                    else {
                        breakpoint = bufpos;
                        break;
                    }
                }
                                    
                // Close if this was the last bytes
                if (clen == -1) {
                    fis.close();
                    closed = true;
                }         
                
                // EOF as breakpoint
                if (breakpoint == -1) {
                    breakpoint = bufpos;
                }
                
                // Inclue breakpoint
                breakpoint++;
                
                // Create chunk
                chunkDigest.update(buf, 0, breakpoint);
                
                byte[] chunkChecksum = chunkDigest.digest();
                byte[] chunkContents = buf;
                int chunkSize = breakpoint;
                byte[] fileChecksum = (clen == -1) ? fileDigest.digest() : null;
                
                return new Chunk(chunkChecksum, chunkContents, chunkSize, fileChecksum);
            } 
            catch (IOException ex) {                
                logger.log(Level.SEVERE, "Error while retrieving next chunk.", ex);
                return null;
            }
        }        
    }    
}

