/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.index.requests;

import java.io.IOException;
import java.util.Enumeration;
import org.syncany.config.Folder;
import org.syncany.db.CloneFile;
import java.io.File;
import java.io.FileOutputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.concurrent.Semaphore;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.zip.GZIPOutputStream;
import javax.crypto.Cipher;
import javax.crypto.CipherOutputStream;
import javax.persistence.EntityManager;
import org.syncany.Constants;
import org.syncany.config.Config;
import org.syncany.db.CloneChunk;
import org.syncany.db.CloneFile.Status;
import org.syncany.db.Database;
import org.syncany.index.chunk.Chunk;
import org.syncany.index.chunk.MetaChunk;
import org.syncany.index.chunk.TarMetaChunk;
import org.syncany.util.FileUtil;

/**
 *
 * @author Philipp C. Heckel
 */
public class NewIndexRequest extends SingleRootIndexRequest {
    private static final Logger logger = Logger.getLogger(NewIndexRequest.class.getSimpleName());
    private static final long METACHUNK_UNFINISHED_TIMEOUT = 1000;
    private static final Semaphore metaChunkMutex = new Semaphore(1);
    
    private static MetaChunk metaChunk;
    private static File metaChunkTempFile;
    private static Set<CloneFile> metaChunkFiles;
    private static EntityManager metaChunkEntityManager;
    
    private static Timer metaChunkTimeoutTimer = new Timer();
    private static TimerTask metaChunkTimoutTimerTask;    

    private File file;
    private CloneFile previousVersion;    

    public NewIndexRequest(Folder root, File file, CloneFile previousVersion) {
        super(root);
        
        this.file = file;        
        this.previousVersion = previousVersion;        
    }

    @Override
    public void process() {
        // .ignore file
        if (file.getName().startsWith(Constants.FILE_IGNORE_PREFIX)) {
            return;
        }
        
        if (logger.isLoggable(Level.INFO)) {
            logger.log(Level.INFO, "Indexer: Indexing new file {0} ...", file);
        }
                
        // File vanished
        if (!file.exists()) {
            if (logger.isLoggable(Level.WARNING)) {
                logger.log(Level.WARNING, "Indexer: Error indexing file {0}: File does NOT exist. Ignoring.", file);
            }
            
            return;
        }
        
        // Create DB entry
        CloneFile newVersion = (previousVersion == null) ? addNewVersion() : addChangedVersion();                      
        
        // Process folder and files differently
        if (file.isDirectory()) {
            processFolder(newVersion);
        }
        
        else if (file.isFile()) {
            processFile(newVersion);
        }
    }

    private CloneFile addNewVersion() {
        CloneFile newVersion = new CloneFile(root, file);        
        
        newVersion.setVersion(1);
        newVersion.setSyncStatus(CloneFile.SyncStatus.LOCAL);
        newVersion.setStatus(Status.NEW);
        newVersion.setUpdated(new Date());
        newVersion.setClientName(config.getMachineName());
        
        return newVersion;
    }
    
    private CloneFile addChangedVersion() {        
        CloneFile newVersion = (CloneFile) previousVersion.clone();

        newVersion.setVersion(previousVersion.getVersion()+1);
        newVersion.setUpdated(new Date());
        newVersion.setStatus(Status.CHANGED);
        newVersion.setSyncStatus(CloneFile.SyncStatus.LOCAL);
        newVersion.setClientName(config.getMachineName());
        newVersion.setChunks(new ArrayList<CloneChunk>()); // otherwise we would append!
        newVersion.setFileSize(file.length());
        newVersion.setLastModified(new Date(file.lastModified()));      
        
        return newVersion;
    }
    
    private void processFolder(CloneFile cf) {        
        // Add rest of the DB stuff 
        cf.setChecksum(null);
        cf.persist();
        
        // Analyze file tree (if directory) -- RECURSIVELY!!
        if (logger.isLoggable(Level.INFO)) {
            logger.log(Level.INFO, "Indexer: Indexing CHILDREN OF {0} ...", file);
        }

        for (File child : file.listFiles()) {
            // Ignore .ignore files
            if (child.getName().startsWith(Constants.FILE_IGNORE_PREFIX)) {                
                continue; 
            }

            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "Indexer: Parent: {0} / CHILD {1} ...", new Object[]{file, child});
            }
            
            // Do it!
            new NewIndexRequest(root, child, null).process();
        }
        
    }
    
    private void processFile(CloneFile cf) {
        try {
            // MUTEX
            metaChunkMutex.acquire();
            
            // 1. Chunk it!
            Chunk chunk = null;
            Enumeration<Chunk> chunks = 
                    root.getProfile().getRepository().getChunker().createChunks(file);
            
            metaChunkEntityManager = config.getDatabase().getEntityManager();   
            
            while (chunks.hasMoreElements()) {
                chunk = chunks.nextElement();      
                
                if (metaChunk != null && metaChunk.isFull()) {
                    closeMetaChunk();
                }
                
                if (metaChunk == null) {
                    byte[] metaId = chunk.getChecksum(); // by convention, the first chunk ID is the metaId
                    int chunkSize = root.getProfile().getRepository().getChunkSize()*1024;

                    Cipher encCipher = root.getProfile().getRepository().getEncryption()
                            .createEncCipher(metaId);

                    metaChunkTempFile = config.getCache().createTempFile("metachunk");
                    metaChunk = new TarMetaChunk(metaId, chunkSize,
                            new CipherOutputStream(new FileOutputStream(metaChunkTempFile), encCipher));
                            //new FileOutputStream(metaChunkTempFile));                    
            //                            new GZIPOutputStream(new FileOutputStream(metaChunkTempFile)));                    
                    metaChunkFiles = new HashSet<CloneFile>();

                    logger.info("Opened new metachunk "+config.getCache().getMetaChunkFile(metaChunk));
    
                    metaChunkEntityManager.getTransaction().begin();
                }     
                

                CloneChunk cc = Database.getInstance().getChunk(metaChunk.getId(), chunk.getChecksum(), true);
                /*CloneChunk cc = new CloneChunk();
                cc.setChecksum(chunk1.getChecksum());
                cc.setMetaId(metaChunk.getId());*/
                //XXXXXXXXXXXXXXxx
                cf.addChunk(cc);
                
                metaChunk.write(chunk);
            }
            
            // 2. Add the rest to the DB, and persist it
            if (chunk != null) {
                // The last chunk holds the file checksum
                cf.setChecksum(chunk.getFileChecksum());                 
            }

            metaChunkFiles.add(cf);
            
            
            // Reset timer
            if (metaChunkTimoutTimerTask != null) {
                metaChunkTimoutTimerTask.cancel();
            }
            
            metaChunkTimoutTimerTask = new MetaChunkTimeoutTimerTask();
            metaChunkTimeoutTimer.schedule(metaChunkTimoutTimerTask, METACHUNK_UNFINISHED_TIMEOUT);            

            // RELEASE MUTEX            
            metaChunkMutex.release();
            
        }  
        catch (Exception e) {
            if (logger.isLoggable(Level.WARNING)) {
                logger.log(Level.WARNING, "Could not index new file "+file+". IGNORING.", e);
            }
            
            return;            
        }
     
           //XXXXXXXXXXXXXXXXXXX unfinished files?
         /*   
            // 1. Chunk it!
            Chunk chunkInfo = null;
            Enumeration<Chunk> chunks = 
                    root.getProfile().getRepository().getChunker().createChunks(file);

            while (chunks.hasMoreElements()) {
                chunkInfo = chunks.nextElement();

                // create chunk in DB (or retrieve it)
                CloneChunk chunk = db.getChunk(chunkInfo.getChecksum(), true);
                cf.addChunk(chunk); 

                // write encrypted chunk (if it does not exist)
                File chunkCacheFile = config.getCache().getCacheChunk(chunk);

                if (!chunkCacheFile.exists()) {
                    byte[] packed = FileUtil.pack(chunkInfo.getContents(), 
                        root.getProfile().getRepository().getEncryption());
                    
                    FileUtil.writeFile(packed, chunkCacheFile);                   
                }
            }
            */
    }
    
    @Override
    public String toString() {
        return NewIndexRequest.class.getSimpleName()+
            "[" + "file=" + file + "]";
    }
    
    private static void closeMetaChunk() throws IOException {        
        Config config = Config.getInstance();
        logger.info("Metachunk "+config.getCache().getMetaChunkFile(metaChunk)+" is FULL.");

        // Close meta chunk and get ID (= last chunk's ID)
        metaChunk.close();

        // Rename to final 'temp' metachunk
        metaChunkTempFile.renameTo(
                config.getCache().getMetaChunkFile(metaChunk));

        for (CloneFile cf2 : metaChunkFiles) {
            metaChunkEntityManager.persist(cf2);
        }
        
        metaChunkEntityManager.flush();
        metaChunkEntityManager.getTransaction().commit();

                    
        for (CloneFile cf2 : metaChunkFiles) {
            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "Indexer: Added to DB. Now Q file {0} at uploader ...", cf2.getFile());
            }            
            
            cf2.getProfile().getUploader().queue(cf2);                            
        }

        //metaChunkEntityManager.getTransaction().begin();

        // Reset meta chunk
        metaChunk = null;           
    }
    
    private class MetaChunkTimeoutTimerTask extends TimerTask {
        @Override
        public void run() {
            try {
                metaChunkMutex.acquire();                
                closeMetaChunk();
                metaChunkMutex.release();
            }
            catch (Exception ex) {
                Logger.getLogger(NewIndexRequest.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
        
    }
    
}
