/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.notification;

import java.util.HashSet;
import java.util.Set;
import org.jivesoftware.smack.Connection;
import org.jivesoftware.smack.PacketListener;
import org.jivesoftware.smack.SASLAuthentication;
import org.jivesoftware.smack.XMPPConnection;
import org.jivesoftware.smack.XMPPException;
import org.jivesoftware.smack.packet.Message;
import org.jivesoftware.smack.packet.Packet;

/**
 *
 * @author Philipp C. Heckel <philipp.heckel@gmail.com>
 */
public class JabberNotification {
    private static final String RESOURCE = "syncany";
    private static final String CMD_NOTIFY = "NOTIFY";
    
    private Connection connection;
    private String host;
    private int port;
    private String username;
    private String password;
    private NotificationListener listener;
    private Set<String> onlineClients;
    
    public JabberNotification() {
        onlineClients = new HashSet<String>();
        port = 5222;        
    }

    public String getHost() {
        return host;
    }

    public void setHost(String host) {
        this.host = host;
    }

    public int getPort() {
        return port;
    }

    public void setPort(int port) {
        this.port = port;
    }

    public String getPassword() {
        return password;
    }

    public void setPassword(String password) {
        this.password = password;
    }

    public String getUsername() {
        return username;
    }

    public void setUsername(String username) {
        this.username = username;
    }

    public NotificationListener getListener() {
        return listener;
    }

    public void setListener(NotificationListener listener) {
        this.listener = listener;
    }
        
    public synchronized void connect() throws XMPPException {
        if (connection != null && connection.isConnected()) {
            return;
        }
        
        for (int i=0; i<3; i++) {
            connection = new XMPPConnection(host);    

            // For Google Talk only?!
            if (host.equals("gmail.com")) {
                SASLAuthentication.supportSASLMechanism("PLAIN", 0);
            }

            connection.connect();
            System.out.println("connected");

            connection.login(username, password, RESOURCE);
            System.out.println("logged in");
            
            listen();
            return;
        }
    }
    
    private void listen() {
        if (!connection.getChatManager().getChatListeners().isEmpty()) {
            return;
        }        
        
        connection.addPacketListener(new PacketListener() {
            @Override
            public void processPacket(Packet packet) {      
                // Ignore messages from myself, ignore Syncany unrelated messages                
                if (packet.getFrom().startsWith(connection.getUser())
                        || !packet.getFrom().contains("/"+RESOURCE)) {
                    
                    return;
                }

                // Add unknown Syncany clients
                if (!onlineClients.contains(packet.getFrom())) {
                    System.out.println(connection.getUser() + " --NEW syncany client found: "+packet.getFrom());
                    onlineClients.add(packet.getFrom());                        
                }
                
                // Ignore everything but 'Message' packets
                if (!(packet instanceof Message)) {
                    return;
                }
                
                // Handle Syncany messages
                Message msg = (Message) packet;
                //System.out.println(connection.getUser()+" -- PACKET:" +packet.toXML());
                
                if (CMD_NOTIFY.equals(msg.getBody())) {
                    System.err.println("Received NOTIFY message: "+msg.toXML());

                    if (listener != null) {
                        listener.notificationReceived(msg.getFrom());
                    }   

                    return;
                }
                else {
                    System.err.println("Received unknown Syncany message: "+packet.toXML());
                }
            }                
        }, null);
        
    }            
    
    private void sendNotification() throws XMPPException {
        connect();                
        
        for (final String toJID : onlineClients) {
            Message msg = new Message();
            msg.setBody(CMD_NOTIFY);
            msg.setTo(toJID);
            
            System.out.println("sending notify "+msg.toXML());            
            connection.sendPacket(msg);
        }
    }    
    
    public interface NotificationListener {
        public void notificationReceived(String fromClient);
    }
    
    public static void main(String[] args) throws XMPPException, InterruptedException {
        /*
         * NOTE: you must be a 'friend' of the person you are notifying.
         *       can we add friends automatically?
         */
        
        JabberNotification sm = new JabberNotification();
        sm.setHost("gmail.com");
        sm.setPort(5222);        
        sm.setUsername("kv3981@gmail.com");
        sm.setPassword("...");
        
        JabberNotification sm2 = new JabberNotification();
        sm2.setHost("gmail.com");
        sm2.setPort(5222);        
        sm2.setUsername("philipp.heckel@gmail.com");
        sm2.setPassword("...");
        
        sm.connect();
        sm2.connect();
        
        sm2.setListener(new NotificationListener() {
            @Override
            public void notificationReceived(String fromClient) {
                System.out.println("from: "+fromClient+" --- now we should manually trigger the remotewatcher.");
            }
        });     
        
        Thread.sleep(2000);
        
        sm.sendNotification();
        
        while (true) {
            Thread.sleep(1000);
        }                
        
    }


}
