/*
 * Syncany, www.syncany.org
 * Copyright (C) 2011 Philipp C. Heckel <philipp.heckel@gmail.com> 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.syncany.uploader;

import java.io.File;
import java.util.Comparator;
import java.util.Date;
import org.syncany.watch.remote.files.RemoteFile;
import org.syncany.exceptions.StorageException;
import org.syncany.db.CloneChunk;
import org.syncany.db.CloneFile;
import org.syncany.config.Config;
import org.syncany.config.Profile;
import org.syncany.connection.plugins.TransferManager;
import org.syncany.db.CloneFile.SyncStatus;
import org.syncany.gui.desktop.Desktop;
import java.util.Map;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.syncany.db.Database;
import org.syncany.gui.tray.Tray;

/**
 * Represents the remote storage.
 * Processes upload and download requests asynchronously.
 *
 * @author Philipp C. Heckel <philipp.heckel@gmail.com>
 */
public class Uploader {
    private static final int CACHE_FILE_LIST = 60000;
    
    private static final Config config = Config.getInstance();
    private static final Logger logger = Logger.getLogger(Uploader.class.getSimpleName());
    private static final Tray tray = Tray.getInstance();
    private static final Desktop desktop = Desktop.getInstance();    
    
    private Profile profile;
    private TransferManager transfer;
    private PriorityBlockingQueue<CloneFile> queue;
    private Thread worker;

    private Map<String, RemoteFile> fileList;
    private Date cacheLastUpdate;
    
    private UploaderComparator uploaderComparator;
    
    public Uploader(Profile profile) {
        this.profile = profile;
        this.uploaderComparator = new UploaderComparator();
        this.queue = new PriorityBlockingQueue<CloneFile>(11, uploaderComparator);

        this.worker = null; // cmp. method 'start'
    }

    public synchronized void start() {
        if (worker != null)
            return;

        transfer = profile.getRepository().getConnection().createTransferManager();
        
        worker = new Thread(new Worker(), "UploaderWorker");        
        worker.start();
    }

    public synchronized void stop() {
        if (worker == null || worker.isInterrupted())
            return;

        worker.interrupt();
        worker = null;
    }

    public synchronized void queue(CloneFile file) {
        queue.put(file);
    }

    private class Worker implements Runnable {
        @Override
        public void run() {
            try {
                CloneFile file;

                while (null != (file = queue.take())) {
                    processRequest(file);
                }
            }
            catch (InterruptedException iex) {
                iex.printStackTrace();
            }
        }

        private void processRequest(CloneFile file) {
            if (tray.getStatusIcon() != Tray.StatusIcon.UPDATING) {
                tray.setStatusIcon(Tray.StatusIcon.UPDATING);
            }

            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "UploadManager: Uploading file {0} ...", file.getFile());
            }

            // Update DB sync status
            if (!file.isFolder()) {
                file.setSyncStatus(CloneFile.SyncStatus.SYNCING);
                file.merge();

                touch(file, SyncStatus.SYNCING);
            }

            // Get file list (to check if chunks already exist)
            if (cacheLastUpdate == null || fileList == null 
                    || System.currentTimeMillis()-cacheLastUpdate.getTime() > CACHE_FILE_LIST) {
                
                try { 
                    fileList = transfer.list();
                }
                catch (StorageException ex) {
                    logger.log(Level.SEVERE, null, ex);
                    return;
                }
            }
            
            for (CloneChunk chunk : file.getChunks()) {
                // Metachunk or single chunk?
                String chunkIdStr; 
                String chunkFileName; 
                
                if (chunk.getMetaId() == null) {
                    chunkIdStr = CloneChunk.encodeIdStr(chunk.getChecksum());
                    chunkFileName = CloneChunk.getFileName(null, chunk.getChecksum());
                }
                else {
                    chunkIdStr = CloneChunk.encodeIdStr(null, chunk.getMetaId());
                    chunkFileName = CloneChunk.getFileName(chunk.getMetaId(), null);
                    
                }
                
                // Chunk has been uploaded before
                if (fileList.containsKey(chunkFileName)) {
                    if (logger.isLoggable(Level.INFO)) {
                        //logger.log(Level.INFO, "UploadManager: Chunk {0} already uploaded", chunkFileName);
                    }
                    
                    continue;
                }

                // Upload it!
                try {
                    if (logger.isLoggable(Level.INFO)) {
                        logger.log(Level.INFO, "UploadManager: Uploading chunk {0} ...", chunkFileName);
                    }
                    
                    RemoteFile chunkFile = new RemoteFile(chunkFileName);
                    transfer.upload(config.getCache().getCacheChunk(chunkIdStr), chunkFile);
                    
                    // Add to cache
                    fileList.put(chunkFileName, chunkFile);
                } 
                catch (StorageException ex) {
                    logger.log(Level.SEVERE,"UploadManager: Uploading chunk "+chunkFileName+" FAILED !!",ex);
                    return;
                }

            }

            if (logger.isLoggable(Level.INFO)) {
                logger.log(Level.INFO, "UploadManager: File {0} uploaded", file.getFile().getAbsolutePath());
            }

            // Update DB sync status
            file.setSyncStatus(SyncStatus.UPTODATE);
            file.merge();
                        
            touch(file, SyncStatus.UPTODATE);

            if (queue.isEmpty()) {
                tray.setStatusIcon(Tray.StatusIcon.UPTODATE);
            }
        }

        private void touch(CloneFile file, SyncStatus syncStatus) {
            // Touch myself
            desktop.touch(file.getFile());
            
            // Touch parents
            Database db = Database.getInstance();
            
            CloneFile childCF = file;
            CloneFile parentCF;
            
            while (null != (parentCF = db.getParent(childCF))) {
                if (parentCF.getSyncStatus() != syncStatus) {
                    parentCF.setSyncStatus(syncStatus);
                    parentCF.merge();
                    
                    desktop.touch(parentCF.getFile());
                }

                childCF = parentCF;
            }
        }
    }
    
    /**
     * Sorts by size for files > 3MB and by name for the others.
     */
    private class UploaderComparator implements Comparator<CloneFile> {
        @Override
        public int compare(CloneFile c1, CloneFile c2) {
            // Files bigger than 3 MB go to the back
            if (c1.getFileSize() > 3*1024*1024 || c2.getFileSize() > 3*1024*1024) {
                if (c1.getFileSize() == c2.getFileSize()) {
                    return 0;
                }
                else if (c1.getFileSize() < c2.getFileSize()) {
                    return -1;
                }
                else {
                    return 1;
                }
            }
            
            // Smaller files by name
            return c1.getFile().compareTo(c2.getFile());
        }        
    }
}
